#include <unistd.h>
#include <stdint.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>

#include <cameraserver/CameraServer.h>

#include <linux/videodev2.h>
#include <opencv2/core/core.hpp>
#include <opencv2/core/types.hpp>
#include <opencv2/imgcodecs.hpp>
#include <opencv2/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <stdexcept>
#include <thread>
#include <vector>
#include <mutex>
#include <condition_variable>

#include <string>
#include <stdexcept>

#include <sys/time.h>
#include <sys/resource.h>
#include <sched.h>

#include <networktables/NetworkTableInstance.h>
#include <networktables/NetworkTable.h>
#include <networktables/DoubleArrayTopic.h>
#include <frc/apriltag/AprilTagDetection.h>
#include <frc/apriltag/AprilTagDetector.h>
#include <frc/apriltag/AprilTagPoseEstimator.h>
#include <frc/geometry/Pose3d.h>
#include <units/length.h>
#include <units/angle.h>

using namespace std;
using namespace frc;

#include "2702StereoVision.h"
#include "v4lcam-mp.h"
#include "2702math.h"
#include "2702camera.h"

#include "rxfm.h"

#include "field.h"

using namespace Rebels;

#include "dualbpi.h"

static FieldXfrms s_fx;


#define resU 736
#define resV 480

void DualBPICam::vInitSensor( v4lcam &vc )
{
	vc.bSetAutoExposure( false );
	vc.bSetAutoGain( false );
	vc.bSetExposure( 480 );

	vc.bInitBuffers();

}

DualBPICam::DualBPICam( char const *pszDev1, char const *pszDev2, char const *pszCamFile ) :
	m_vcLeft( pszDev1 ),
	m_vcRight( pszDev2 )
{
	m_vcLeft.bSetSelection( 1, 5, resU, resV );
	m_vcRight.bSetSelection( 1, 5, resU, resV );
	m_vcLeft.bSetRes( resU, resV, V4L2_PIX_FMT_SGRBG8 );
	m_vcRight.bSetRes( resU, resV, V4L2_PIX_FMT_SGRBG8 );

#if defined(MJPEG)
	for( int i = 0; i < 2; i++ ) {
		char szTmp[100];
		sprintf( szTmp, "Raw%d", i );
		m_grcvsOutput[i] = frc::CameraServer::PutVideo( szTmp, resU, resV );
	}
#endif
	vInitSensor( m_vcLeft );
	vInitSensor( m_vcRight );

	if( !m_sc.bLoad( pszCamFile )) {
		throw runtime_error( string( "Unable to load camera parameters" ) );
	}
	auto ntinst = nt::NetworkTableInstance::GetDefault();
	m_dapXYR = ntinst.GetDoubleArrayTopic("/Vision/PoseArray").Publish();

	m_th3DProc = std::thread( &DualBPICam::calc3DThread, this );
	m_thCamera = std::thread( &DualBPICam::cameraThreadFn, this );
}

DualBPICam::DualBPICam( int fdLeft, int fdRight ) :
	m_vcLeft( fdLeft ),
	m_vcRight( fdRight )
{
	vInitSensor( m_vcLeft );
	vInitSensor( m_vcRight );

	m_th3DProc = std::thread( &DualBPICam::calc3DThread, this );
	m_thCamera = std::thread( &DualBPICam::cameraThreadFn, this );
}

DualBPICam::~DualBPICam()
{
	m_bShuttingDown = true;
	bQueueFrame( -1 );

	m_th3DProc.join();
	m_thCamera.join();
	if( m_fpRaw )
		fclose( m_fpRaw );
}


bool DualBPICam::bQueueFrame( int nBuffer )
{
	std::unique_lock<std::mutex> l(m_mtxQueue);

	// return false indicating that the 3D Thread was busy
	if( m_nWorkBuff >= 0 )
		return false;

	m_nWorkBuff = nBuffer;
	m_cvQueue.notify_one();
	return true;
}

int DualBPICam::nWaitForFrame()
{
	std::unique_lock<std::mutex>	l(m_mtxQueue);
	m_nWorkBuff = -1;
	m_cvQueue.wait( l, [&] { return (m_nWorkBuff != -1 || m_bShuttingDown); } );

	return m_nWorkBuff;
}

int DualBPICam::nDetectAprilTag( uint8_t const **ppu8Img, int nSensor )
{
	AprilTagDetector       atd;
	atd.AddFamily("tag16h5", 0);

	AprilTagDetector::Results detections = atd.Detect( resU, resV, (uint8_t *)ppu8Img[nSensor] );
#if defined(MJPEG)
	cv::Mat raw(resV, resU, CV_8UC1, (void *)ppu8Img[nSensor] );
#endif
	for (AprilTagDetection const *detection : detections ) {
		int nId = detection->GetId();
		if( nId >= 1 && nId <= 8 ) {
			nId -= 1;
			for( int c = 0; c < 4; c++ ) {
printf( "Sensor %d ID: %d\n", nSensor, nId + 1);
				auto p = detection->GetCorner( c );
				m_grTagInfo[nId].Corner( nSensor, c ).set( p.x, p.y );
				m_grTagInfo[nId].bValid2d( nSensor ) = true;
#if defined(MJPEG)
				cv::Point cp1( (int)p.x, (int)p.y );
				auto &p2 = detection->GetCorner( (c + 1) & 3 );

				cv::Point cp2( (int)(p.x + (p2.x - p.x) * (c+1)/4), (int)(p.y + (p2.y - p.y) * (c+1)/4) );

				cv::line( raw, cp1, cp2, cv::Scalar(255), 2 );
#endif
			}
		}
	}
#if defined(MJPEG)
	m_grcvsOutput[nSensor].PutFrame( raw );
#endif

	return 0;
}

void CvtMatToQuat(Rebels::Quaternion &q, Point3d &p, cv::Mat &m) 
{
	assert( m.rows == 3 && m.cols == 4 );

	float grf[3][3];

	float fQ0 = 1.F;
	float fX = 0.F;
	float fY = 0.F;
	float fZ = 0.F;

	for( int i = 0; i < 3; i++ ) {
		for( int j = 0; j < 3; j++ ) {
			grf[i][j] = m.at<float>(i,j);
		}
	}
	float fT = grf[0][0] + grf[1][1] + grf[2][2] + 1;
	if (fT > 0.00000001F) {
		float fS = sqrt(fT) * 2.F;
		fQ0 = 0.25F * fS;
		fX = (grf[2][1] - grf[1][2]) / fS;
		fY = (grf[0][2] - grf[2][0]) / fS;
		fZ = (grf[1][0] - grf[0][1]) / fS;
	} else if (grf[0][0] > grf[1][1] &&
			 grf[0][0] > grf[2][2]) {
		float fS =
			sqrt(1.0F + grf[0][0] - grf[1][1] - grf[2][2]) *
			2.F;
		fX = 0.25F * fS;
		fY = (grf[1][0] + grf[0][1]) / fS;
		fZ = (grf[0][2] + grf[2][0]) / fS;
		fQ0 = (grf[2][1] - grf[1][2]) / fS;
	} else if (grf[1][1] > grf[2][2]) {
		float fS =
			sqrt(1.0F + grf[1][1] - grf[0][0] - grf[2][2]) *
			2.F;
		fX = (grf[1][0] + grf[0][1]) / fS;
		fY = 0.25F * fS;
		fZ = (grf[2][1] + grf[1][2]) / fS;
		fQ0 = (grf[0][2] - grf[2][0]) / fS;
	} else {
		float fS =
			sqrt(1.0F + grf[2][2] - grf[0][0] - grf[1][1]) *
			2.F;
		fX = (grf[0][2] + grf[2][0]) / fS;
		fY = (grf[2][1] + grf[1][2]) / fS;
		fZ = 0.25F * fS;
		fQ0 = (grf[1][0] - grf[0][1]) / fS;
	}

	q = Rebels::Quaternion( fQ0, fX, fY, fZ );
	p = Point3d( m.at<float>(0,3), m.at<float>(1,3), m.at<float>(2,3) );
}

void DualBPICam::calc3DThread()
{
	Point3d gr3dMeas[32];	// 8 april tags * 4 markers
	Point3d gr3dRef[32];

	while( !m_bShuttingDown )
	{
		int nBuf = nWaitForFrame();

		if( nBuf >= 0 ) {
			int nVisPts = 0;
//			printf( " %u->%u\n", m_grLeftBuffers[nBuf].sequence, m_grRightBuffers[nBuf].sequence );

			uint8_t const *grpBuffers[2];
			grpBuffers[0] = m_vcLeft.pu8GetImage( m_grLeftBuffers[nBuf].index );
			grpBuffers[1] = m_vcRight.pu8GetImage( m_grRightBuffers[nBuf].index );

			// clear out the TagInfo
			for( int i = 0; i < 8; i++ ) {
				m_grTagInfo[i].ClearFrame();
			}

			#pragma omp parallel for
			for( int i = 0; i < 2; i++) {
				nDetectAprilTag( grpBuffers, i );
			}
			
			bool bValid = false;
			for( int nTag = 0; nTag < 8; nTag++ ) {
				AprilTagInfo &ti = m_grTagInfo[nTag];
				if( ti.bValid2d(0) && ti.bValid2d(1) ) {
					for( int nC = 0; nC < 4; nC++ ) {
						gr3dRef[nVisPts] = s_fx[nTag+1][nC];

						float fLineSep = m_sc.fCalc3D( m_grTagInfo[nTag].m_grp2dCorners[0][nC], m_grTagInfo[nTag].m_grp2dCorners[1][nC], gr3dMeas[nVisPts] );
printf( "%d: %8.3f %8.3f %8.3f  %8.3f\n", nC, gr3dMeas[nVisPts].x(), gr3dMeas[nVisPts].y(), gr3dMeas[nVisPts].z(), fLineSep );
						nVisPts++;
					}
					bValid = true;
				}
			}

			if (nVisPts) {
				printf( "Have %d Pts : ", nVisPts );
				cv::Mat_<cv::Vec3f> vMeasPts(nVisPts,1);
				cv::Mat_<cv::Vec3f> vRefPts(nVisPts,1);

				// build ref and measured arrays of points
				for( int nMkr = 0; nMkr < 4; nMkr++ ) {
					vMeasPts.at<cv::Vec3f>(nMkr) = cv::Vec3f( gr3dMeas[nMkr].x(), gr3dMeas[nMkr].y(), gr3dMeas[nMkr].z() );
					vRefPts.at<cv::Vec3f>(nMkr)  = cv::Vec3f( gr3dRef[nMkr].x(), gr3dRef[nMkr].y(), gr3dRef[nMkr].z() );
				}
//				cv::Mat_<float> mRes = FindRigidTransform(vRefPts,vMeasPts);
				cv::Mat_<float> mRes = FindRigidTransform(vMeasPts,vRefPts);
				Rebels::Quaternion qRot;
				Point3d p3dXlat;

				CvtMatToQuat(qRot, p3dXlat, mRes);
				printf( "(%8.4f,%8.4f,%8.4f,%8.4f) (%10.2f,%10.2f,%10.2f)\n", qRot.q0(), qRot.x(), qRot.y(), qRot.z(), p3dXlat.x(), p3dXlat.y(), p3dXlat.z() );
				frc::Quaternion q( qRot.q0(), qRot.x(), qRot.y(), qRot.z() );

//				Transform3d x3d( Translation3d(units::length::inch_t(p3dXlat.x()/25.4), units::length::inch_t(p3dXlat.y()/25.4), units::length::inch_t(p3dXlat.z()/25.4)), Rotation3d(q) );
				Pose3d pose( Translation3d(units::length::inch_t(p3dXlat.x()/25.4), units::length::inch_t(p3dXlat.y()/25.4), units::length::inch_t(p3dXlat.z()/25.4)), Rotation3d(q) );
				
				frc::Pose2d p2d = pose.ToPose2d();

#if 0
				struct timespec tsCur;
				struct timeval tvCur;
				struct timeval tvDelta;

				// Get current time since power on
				clock_gettime(CLOCK_MONOTONIC, &tsCur );
				TIMESPEC_TO_TIMEVAL( &tvCur, &tsCur );

				// Calculate delta between when frame was catured and now.
				timersub( &tvCur, pFrm->ts(), &tvDelta );

				// Get current NetTables time and subtract off the delta
				int64_t time = nt::Now();
				time -= tvDelta.tv_usec;
#else
				int64_t time = nt::Now();
#endif
				printf( "Setting pose to : %8.2f, %8.2f / %8.2f\n", p2d.X().value(), p2d.Y().value(), p2d.Rotation().Degrees().value() );
				m_dapXYR.Set( wpi::SmallVector<double, 9> { p2d.X().value(), p2d.Y().value(), p2d.Rotation().Radians().value() }, time );
			}
			else {
				int64_t time = nt::Now();
				m_dapXYR.Set( wpi::SmallVector<double, 9> { -1., -1., -1. }, time );
			}
			if( bValid && m_fpRaw) {
				for( int nTag = 0; nTag < 8; nTag++ ) {
					AprilTagInfo &ti = m_grTagInfo[nTag];
					if( ti.bValid2d(0) && ti.bValid2d(1) ) {
						fprintf( m_fpRaw, "Tag %d", nTag );
						for( int nSensor = 0; nSensor < 2; nSensor++ ) {
							for( int nCorner = 0; nCorner < 4; nCorner++ ) {
								Point2d &c = ti.Corner( nSensor, nCorner );
								fprintf( m_fpRaw, " (%9.4f,%9.4f)", c.u(), c.v() );
							}
						}
						fprintf( m_fpRaw, "\n" );
					}
				}
			}
			m_vcLeft.FrameComplete( m_grLeftBuffers[nBuf].index );
			m_vcRight.FrameComplete( m_grRightBuffers[nBuf].index );
		}
		else
			printf( ">>>>>\n" );
	}
}

void DualBPICam::cameraThreadFn()
{
	bool bSyncing = false;
	bool bSynced = true;
	int nCurBuf = 0;

	m_vcLeft.bStart();
	m_vcRight.bStart();
	while( !m_bShuttingDown ) {
		m_vcLeft.WaitForFrame( m_grLeftBuffers[nCurBuf] );
		m_vcRight.WaitForFrame( m_grRightBuffers[nCurBuf] );

		if( !bSynced ) {
			auto const &lb = m_grLeftBuffers[nCurBuf];
			auto const &rb = m_grRightBuffers[nCurBuf];
//printf( "Left %d.%06d  Right %d.%06d\n", lb.timestamp.tv_sec, lb.timestamp.tv_usec, rb.timestamp.tv_sec, rb.timestamp.tv_usec );
timeval tv1,tv2;
timersub( &rb.timestamp, &lb.timestamp, &tv1 );
timersub( &lb.timestamp, &rb.timestamp, &tv2 );
printf( "Diff: %u    %u.%06u    %u    %u.%06u\n", lb.sequence, tv1.tv_sec, tv1.tv_usec, rb.sequence, tv2.tv_sec, tv2.tv_usec );
			m_vcLeft.FrameComplete( m_grLeftBuffers[nCurBuf].index );
			m_vcRight.FrameComplete( m_grRightBuffers[nCurBuf].index );
		}
		else
		{
			// Queue the Frame for 3D processing.
			//   If the frame was sent for 3D processing, use the next buffer
			if( bQueueFrame( nCurBuf ) ) {
				nCurBuf = (nCurBuf + 1) % 2;
			}
			else { // 3D processing task was busy
				m_vcLeft.FrameComplete( m_grLeftBuffers[nCurBuf].index );
				m_vcRight.FrameComplete( m_grRightBuffers[nCurBuf].index );
			}
		}

	}
	m_vcLeft.bStop();
	m_vcRight.bStop();
}

void DualBPICam::SetRawLog( char const *psz )
{
	FILE *fp = fopen( psz, "w" );
	if( fp )
		m_fpRaw = fp;
}
